 /*
 * (c) Copyright 2015 -- Anders Torger
 *
 * This program is open source. For license terms, see the LICENSE file.
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <assert.h>

#include <spectraldb.h>
#include <observers.h>
#include <interp.h>
#include <look.h>
#include <lut.h>
#include <dngref.h>
#include <icclut.h>

struct {
    int locus_poly_count;
    double *locus_poly_x;
    double *locus_poly_y;
} glob;

static v3
icclut_iccv2clip(v3 lab)
{
    if (!isfinite(lab.v[0]) || !isfinite(lab.v[1]) || !isfinite(lab.v[2])) {
        v3_print("bad lab", lab);
        abort();
    }
    // limit to the maximum representable values in ICCv2
    if (lab.v[0] < 0) lab.v[0] = 0;
    if (lab.v[0] > 100 + 25500/65280.0) lab.v[0] = 100 + 25500/65280.0;
    v3 lch = lab2lch(lab);
    while (lab.v[1] < -128.0  || lab.v[1] > 127 + 255/256.0 || lab.v[2] < -128.0  || lab.v[2] > 127 + 255/256.0) {
        double h = lch.v[2] / 180.0 * M_PI;
        if (lab.v[1] < -128.0) {
            if (lab.v[1] > -129.0) {
                lab.v[1] = -128.0;
            } else {
                lch.v[1] = -128.0 / cos(h);
                lab = lch2lab(lch);
            }
        } else if (lab.v[1] > 127 + 255/256.0) {
            if (lab.v[1] < 129.0) {
                lab.v[1] = 127 + 255/256.0;
            } else {
                lch.v[1] = (127 + 255/256.0) / cos(h);
                lab = lch2lab(lch);
            }
        } else if (lab.v[2] < -128.0) {
            if (lab.v[2] > -129.0) {
                lab.v[2] = -128.0;
            } else {
                lch.v[1] = -128.0 / sin(h);
                lab = lch2lab(lch);
            }
        } else if (lab.v[2] > 127 + 255/256.0) {
            if (lab.v[2] < 129.0) {
                lab.v[2] = 127 + 255/256.0;
            } else {
                lch.v[1] = (127 + 255/256.0) / sin(h);
                lab = lch2lab(lch);
            }
        }
    }
    return lab;
}

static bool
gamut_isvalid(v3 xyz)
{
    if (xyz.v[0] == 0 && xyz.v[1] == 0 && xyz.v[2] == 0) {
        return true;
    }
    if (xyz.v[0] < 0 || xyz.v[1] < 0 || xyz.v[2] < 0) {
        return false;
    }
    if (!isfinite(xyz.v[0]) || !isfinite(xyz.v[1]) || !isfinite(xyz.v[2])) {
        return false;
    }
    v3 xyY = xyz2xyY(xyz);
    if (point_inside_polygon(glob.locus_poly_x, glob.locus_poly_y, glob.locus_poly_count, xyY.v[0], xyY.v[1])) {
        return true;
    }
    return false;
}

static double
cam_similarity(v3 cam1,
               v3 cam2)
{
    double max1 = v3_max(cam1);
    double max2 = v3_max(cam2);
    assert(max1 != 0);
    assert(max2 != 0);
    cam1 = k_mul_v3(1.0 / max1, cam1);
    cam2 = k_mul_v3(1.0 / max2, cam2);

    double d = euclidean_dist(cam1, cam2);

    return d;
}

static inline double
icclut_incurve(double x,
               bool invert)
{
    if (!invert) {
        const double p = pow(3.0, 9/2.0);
        const double t = 1.0/3.0;
        return pow(sqrt(6075*x*x+1)/p+5*x/9, t) - 1/(27*pow(sqrt(6075*x*x+1)/p+5*x/9, t));
    } else {
        return 0.1*x+0.9*pow(x,3);
    }
}

struct icc_lut *
icclut_new(const m3x3 cam2xyz,
           const v3 prof_wp,
           const v3 pcs_wp,
           const v3 wb,
           double *trc[3],
           int trc_count,
           const double tc[],
           int tc_len,
           enum tc_type tc_type,
           bool skip_tc_apply,
           chromalut_t *prof_lut,
           bool pcs_is_lab,
           int clut_side,
           const struct tone_rep_op_config *ntro_conf)
{
    const int cs = clut_side;
    struct icc_lut *lut = calloc(1, sizeof(*lut) + 3*(256 + 2 + cs*cs*cs) * sizeof(lut->data[0]));
    double *pd = lut->data;

    if (clut_side > 255) {
        fprintf(stderr, "Error: 3D LUT side may not be larger than 255.\n");
        exit(EXIT_FAILURE);
    }

    lut->out_len = 2;
    lut->in_len = 256;
    lut->clut_side = cs;
    for (int i = 0; i < 3; i++) {
        lut->out[i] = pd;
        pd += lut->out_len;
        lut->in[i] = pd;
        pd += lut->in_len;
    }
    lut->clut = pd;

    for (int i = 0; i < 3; i++) {
        lut->out[i][0] = 0;
        lut->out[i][1] = 1;
        for (int j = 0; j < lut->in_len; j++) {
            double x = (double)j / (lut->in_len-1);
            lut->in[i][j] = icclut_incurve(x, false);
        }
    }
    const m3x3 xyz2rgb = dcp_xyz_D50_to_prophoto_rgb;
    const m3x3 rgb2xyz = invert_m3x3(xyz2rgb);
    look_tone_rep_op_t *ntro = NULL;
    const int itc_len = 65535;
    double *itc = NULL;
    if (tc != NULL) {
        if (tc_type == TC_NEUTRAL) {
            ntro = look_tone_rep_op_new(tc, tc_len, ntro_conf);
            itc = malloc(itc_len * sizeof(itc[0]));
            for (int i = 0; i < itc_len; i++) {
                double x = (double)i / (itc_len - 1);
                itc[i] = inverse_curve(x, tc, tc_len);
            }
        }
    }

    if (pcs_is_lab) {
        lut->cm.v[0][0] = lut->cm.v[1][1] = lut->cm.v[2][2] = 1.0;
    } else {
        lut->cm = cam2xyz;
        v3 wb_mul = v3_norm(v3_invert(wb));
        lut->cm = m3x3_mul_m3x3(lut->cm, v3_asdiagonal(wb_mul));
    }
    // normally it should be tiny differences between pcs_wp and prof_wp
    m3x3 map_forward = bradford_map_white(pcs_wp, prof_wp);
    m3x3 map_reverse = bradford_map_white(prof_wp, pcs_wp);
    v3 wb_mul = v3_norm(v3_invert(wb));
    fprintf(stderr, "Generating ICC RGB to %s LUT with %dx%dx%d = %d entries...\n  ", pcs_is_lab ? "Lab" : "XYZ", cs, cs, cs, cs*cs*cs);
    int complete_count, last_percent;

    fprintf(stderr, "Calculating valid in-gamut entries...");
    v3 *cammap = malloc(cs * cs * cs * sizeof(cammap[0]));
    uint8_t *offsets = malloc(3 * cs * cs * cs * sizeof(offsets[0]));
    int offsets_len = 0;
    uint8_t *ig_offsets = malloc(3 * cs * cs * cs * sizeof(ig_offsets[0]));
    int ig_offsets_len = 0;
    complete_count = last_percent = 0;
#pragma omp parallel for
    for (int r_ = 0; r_ < cs; r_++) {
#pragma omp critical
        {
            fprintf(stderr, ".");
            int percent = (100 * complete_count) / cs;
            if (percent < 100 && percent >= last_percent + 10) {
                fprintf(stderr, "%d%%", percent);
                last_percent = percent;
            }
            complete_count++;
        }
        for (int g_ = 0; g_ < cs; g_++) {
            for (int b_ = 0; b_ < cs; b_++) {
                double r = (double)r_ / (cs-1);
                double g = (double)g_ / (cs-1);
                double b = (double)b_ / (cs-1);
                v3 cam = {{ r, g, b}};
                for (int i = 0; i < 3; i++) {
                    cam.v[i] = icclut_incurve(cam.v[i], true);
                    if (pcs_is_lab) {
                        cam.v[i] *= wb_mul.v[i];
                    }
                }
                if (trc_count > 0) {
                    for (int i = 0; i < 3; i++) {
                        double v = cam.v[i];
                        if (v < 0) v = 0;
                        if (v > 1) v = 1;
                        if (trc_count == 1) {
                            cam.v[i] = pow(cam.v[i], trc[i][0]);
                        } else {
                            int idx = floor(v * (trc_count-1));
                            double d = v * (trc_count-1) - idx;
                            if (d > 1e-7) {
                                cam.v[i] = (1.0 - d) * trc[i][idx] + d * trc[i][idx+1];
                            } else {
                                cam.v[i] = trc[i][idx];
                            }
                        }
                    }
                }
                v3 cxyz = pcs_is_lab ? m3x3_mul_v3(cam2xyz, cam) : cam;
                if (prof_lut != NULL) {
                    cxyz = m3x3_mul_v3(map_forward, cxyz);
                    cxyz = chromalut_lookup_noclip(prof_lut, cxyz);
                    cxyz = m3x3_mul_v3(map_reverse, cxyz);
                }
                bool in_gamut = gamut_isvalid(cxyz);
                if (in_gamut) {
                    if (ntro != NULL) {
                        v3 tcxyz;
                        if (look_tone_rep_op(&tcxyz, rgb2xyz, xyz2rgb, cxyz, ntro)) {
                            cxyz = tcxyz;
                        } else {
                            in_gamut = false;
                        }
                        if (skip_tc_apply) {
                            v3 rgb = v3_clip01(m3x3_mul_v3(xyz2rgb, cxyz));
                            rgb.v[0] = curve_y(rgb.v[0], itc, itc_len);
                            rgb.v[1] = curve_y(rgb.v[1], itc, itc_len);
                            rgb.v[2] = curve_y(rgb.v[2], itc, itc_len);
                            cxyz = m3x3_mul_v3(rgb2xyz, rgb);
                        }
                    } else if (tc != NULL && !skip_tc_apply) {
                        v3 rgb = v3_clip01(m3x3_mul_v3(xyz2rgb, cxyz));
                        rgb.v[0] = curve_y(rgb.v[0], tc, tc_len);
                        rgb.v[1] = curve_y(rgb.v[1], tc, tc_len);
                        rgb.v[2] = curve_y(rgb.v[2], tc, tc_len);
                        cxyz = m3x3_mul_v3(rgb2xyz, rgb);
                    }
                    if (in_gamut) {
                        in_gamut = gamut_isvalid(cxyz);
                    }
                }
                if (!in_gamut) {
#pragma omp critical
                    {
                        offsets[3*offsets_len+0] = r_;
                        offsets[3*offsets_len+1] = g_;
                        offsets[3*offsets_len+2] = b_;
                        offsets_len++;
                    }
                } else {
#pragma omp critical
                    {
                        if (r_ != 0 || g_ != 0 || b_ != 0) {
                            ig_offsets[3*ig_offsets_len+0] = r_;
                            ig_offsets[3*ig_offsets_len+1] = g_;
                            ig_offsets[3*ig_offsets_len+2] = b_;
                            ig_offsets_len++;
                        }
                    }
                }
                cammap[cs*cs*r_+cs*g_+b_] = cam;
                for (int i = 0; i < 3; i++) {
                    lut->clut[3*(cs*cs*r_+cs*g_+b_)+i] = cxyz.v[i];
                }
            }
        }
    }
    fprintf(stderr, "..100%%\n");

    {
        fprintf(stderr, "  Interpolating out of gamut entries...");
        complete_count = last_percent = 0;
#pragma omp parallel for
        for (int i = 0; i < offsets_len; i++) {
#pragma omp critical
            {
                int percent = (100 * complete_count) / offsets_len;
                if (percent < 100 && percent >= last_percent + 10) {
                    fprintf(stderr, "..%d%%", percent);
                    last_percent = percent;
                }
                complete_count++;
            }
            const int r_ = offsets[3*i+0];
            const int g_ = offsets[3*i+1];
            const int b_ = offsets[3*i+2];
            double min_dist = 10000000;
            v3 cam = cammap[cs*cs*r_+cs*g_+b_];
            double *da = malloc(ig_offsets_len * sizeof(da[0]));
            for (int j = 0; j < ig_offsets_len; j++) {
                const int r1 = ig_offsets[3*j+0];
                const int g1 = ig_offsets[3*j+1];
                const int b1 = ig_offsets[3*j+2];
                double d = cam_similarity(cam, cammap[cs*cs*r1+cs*g1+b1]);
                da[j] = d;
                if (d < min_dist) {
                    min_dist = d;
                }
            }
            int m_count = 0;
            v3 cam_sum={{0,0,0}};
            v3 xyz_sum={{0,0,0}};
            min_dist *= 1.2;
            for (int j = 0; j < ig_offsets_len; j++) {
                if (da[j] <= min_dist) {
                    const int r1 = ig_offsets[3*j+0];
                    const int g1 = ig_offsets[3*j+1];
                    const int b1 = ig_offsets[3*j+2];
                    v3 cam1 = cammap[cs*cs*r1+cs*g1+b1];
                    for (int k = 0; k < 3; k++) {
                        cam_sum.v[k] += cam1.v[k];
                        xyz_sum.v[k] += lut->clut[3*(cs*cs*r1+cs*g1+b1)+k];
                    }
                    m_count++;
                }
            }
            free(da);
            cam_sum = k_mul_v3(1.0 / m_count, cam_sum);
            xyz_sum = k_mul_v3(1.0 / m_count, xyz_sum);
            double lum1 = v3_max(cam);
            double lum2 = v3_max(cam_sum);
            xyz_sum = k_mul_v3(lum1 / lum2, xyz_sum);
            double mx = v3_max(xyz_sum);
            if (mx > 1.0) {
                xyz_sum = k_mul_v3(1.0 / mx, xyz_sum);
            }
            for (int i = 0; i < 3; i++) {
                lut->clut[3*(cs*cs*r_+cs*g_+b_)+i] = xyz_sum.v[i];
            }
        }
        fprintf(stderr, "..100%%\n");
    }
    free(ig_offsets);

    free(cammap);
    free(offsets);

    if (pcs_is_lab) {
        fprintf(stderr, "  Converting from XYZ to Lab...");

        { // convert to lab
#pragma omp parallel for
            for (int r_ = 0; r_ < cs; r_++) {
                for (int g_ = 0; g_ < cs; g_++) {
                    for (int b_ = 0; b_ < cs; b_++) {
                        v3 cxyz = {{ lut->clut[3*(cs*cs*r_+cs*g_+b_)+0], lut->clut[3*(cs*cs*r_+cs*g_+b_)+1], lut->clut[3*(cs*cs*r_+cs*g_+b_)+2] }};
                        v3 lab = xyz2lab(cxyz, pcs_wp);
                        lab = icclut_iccv2clip(lab);
                        lut->clut[3*(cs*cs*r_+cs*g_+b_)+0] = lab.v[0];
                        lut->clut[3*(cs*cs*r_+cs*g_+b_)+1] = lab.v[1];
                        lut->clut[3*(cs*cs*r_+cs*g_+b_)+2] = lab.v[2];
                    }
                }
            }
        }

        {
            // map 0,0,0 to exactly 0,0,0
            lut->clut[0] = lut->clut[1] = lut->clut[2] = 0;

            // if chromaticity of 1,1,1 is close to zero, set to zero
            int offset = 3*(cs*cs*(cs-1)+cs*(cs-1)+cs-1);
            for (int i = 0; i < 2; i++) {
                double ov = lut->clut[offset+1+i];
                if (fabs(ov) < 0.01) {
                    ov = 0;
                }
                lut->clut[offset+1+i] = ov;
            }
            // if lightness of 1,1,1 is close to 100 set to 100
            if (fabs(lut->clut[offset+0] - 100.0) < 0.01) {
                lut->clut[offset+0] = 100.0;
            }
        }

        fprintf(stderr, "done!\n");
    }
    look_tone_rep_op_delete(ntro);
    free(itc);
    fprintf(stderr, "done!\n");
    return lut;
}

void
icclut_init(void)
{
    const struct observer *obs = observer_get(OBSERVER_1931_2);
    const spectrum_t *illuminant = spectraldb_illuminant(lsD50);

    const double peak[3] = { 0, 1, 0 };
    int start = spec_band_start(obs->cmf[0]);
    int end = spec_band_end(obs->cmf[0]);
    int i;
    for (i = start; i <= end-10; i += 5) {
        spectrum_t *s = spec_alloc(peak, i, i+10, 5);
        v3 xyz = spec2xyz_ill(s, obs, illuminant, 1);
        free(s);
        if (xyz.v[1] >= 0.003) {
            break;
        }
    }
    start = i;
    for (i = end-10; i >= start; i -= 5) {
        spectrum_t *s = spec_alloc(peak, i, i+10, 5);
        v3 xyz = spec2xyz_ill(s, obs, illuminant, 1);
        free(s);
        if (xyz.v[1] >= 0.003) {
            break;
        }
    }
    end = i;
    const int lop_count = 20;
    int poly_count = (end - start) / 5 + 1 + lop_count;
    double *locus_poly_x = malloc(poly_count * sizeof(*locus_poly_x));
    double *locus_poly_y = malloc(poly_count * sizeof(*locus_poly_y));
    for (int i = 0; i < poly_count; i++) {
        spectrum_t *s;
        if (i < poly_count - lop_count) {
            s = spec_alloc(peak, start+i*5, start+i*5+10, 5);
        } else {
            // line of purples
            double freq[6] = { start, start+5, start+10,  end, end+5, end+10 };
            int k = i - (poly_count - lop_count);
            double a = (double)(k+1) / (lop_count+1);
            a = pow(a, 2);
            double dual_peak[6] = { 0,a,0, 0,1-a,0 };
            s = spec_alloc1(freq, dual_peak, 6);
        }
        v3 xyz = spec2xyz_ill(s, obs, illuminant, 1);
        v3 xyY = xyz2xyY(xyz);
        locus_poly_x[i] = xyY.v[0];
        locus_poly_y[i] = xyY.v[1];
        //v3 ltu = xyz2lutspace(xyz);
        //printf("%f %f %f\n", ltu.v[1], ltu.v[2], ltu.v[0]);
        free(s);
    }
    glob.locus_poly_x = locus_poly_x;
    glob.locus_poly_y = locus_poly_y;
    glob.locus_poly_count = poly_count;
}
